"""
-----------------------------------------------------------------------------------------
bitlevel module
-----------------------------------------------------------------------------------------
Provides functions for exploring bit-level representations of images. 
"""

import numpy as np


def bit_quantization(img: np.ndarray, num_bits: int = 8) -> np.ndarray:
    """Simulates the effect of bit-quantization on an image.

    Args:
        img (np.ndarray): Imput image.

        num_bits (int, optional): Number of bits to simulate quantization.
        Defaults to 8.

    Returns:
        np.ndarray: Output image.
    """
    return ((img >> (8 - num_bits)) / (2**num_bits)) * (2**8 - 1)


def bit_plane(img: np.ndarray, bit: int = 8) -> np.ndarray:
    """Get the nth bit plane.

    Args:
        img (np.ndarray): Imput image.

        bit (int, optional): Bit plane to obtain.
        Defaults to 8.

    Returns:
        np.ndarray: Output image.
    """
    return ((img >> bit) % 2) * (2**8 - 1)



"""
-----------------------------------------------------------------------------------------
pointprocessing module
-----------------------------------------------------------------------------------------
Provides point processing functionality. 
"""

import matplotlib.pyplot as plt
import numpy as np
import cv2 as cv


def plot_transformation(
    x: np.ndarray,
    y: np.ndarray,
    fig_title: str = "Transformation",
    save_figure: bool = False,
    save_name: str = "Figure.jpg",
    save_format: str = "jpg",
) -> None:
    """Visualizes pixel transformation curves.

    Args:
        x (np.ndarray): Input intensity values.

        y (np.ndarray): Output intensity values.

        fig_title (str, optional): Figure title.
        Defaults to "Transformation".

        save_figure (bool, optional): To save figure.
        Defaults to False.

        save_name (str, optional): File name.
        Defaults to "Figure.jpg".

        save_format (str, optional): File extension.
        Defaults to "jpg".
    """

    fig = plt.figure(figsize=(5, 5))
    ax = fig.gca()

    ax.plot(x, y)
    ax.set_xticks([0, 128, 255])
    ax.set_yticks([0, 128, 255])
    ax.set_xlim((0, 256))
    ax.set_ylim((0, 256))

    ax.set_xlabel("input intensity", fontsize="x-large")
    ax.set_ylabel("output intensity", fontsize="x-large")
    ax.tick_params(axis="both", labelsize="x-large")  # Increase size of tick labels

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    ax.set_title(fig_title, fontsize="xx-large")

    plt.show()

    if save_figure:
        fig.savefig(save_name, dpi=300, format=save_format, bbox_inches="tight")


def linear_norm(img: np.ndarray, v_min: int = 0, v_max: int = 255) -> np.ndarray:
    """Normalize an image.

    Args:
        img (np.ndarray): Imput image.

        v_min (int, optional): Minimum intensity value.
        Defaults to 0.

        v_max (int, optional): Maximum intensity value.
        Defaults to 255.

    Returns:
        np.ndarray: Normalized image.
    """

    return np.round(
        (img - img.min()) * ((v_max - v_min) / (img.max() - img.min())) + v_min
    )


def intensity_clamp(img: np.ndarray, v_min: int = 0, v_max: int = 255) -> np.ndarray:
    """Limits the output values of an image.

    Args:
        img (np.ndarray): Input image.

        v_min (int, optional): Minimum allowed intensity level.
        Defaults to 0.

        v_max (int, optional): Maximum allowed intensity level.
        Defaults to 255.

    Returns:
        np.ndarray: Output image.
    """
    img_c = img.copy()
    img_c[img_c < v_min] = v_min
    img_c[img_c > v_max] = v_max

    return img_c


def unsharp_masking(
    img: np.ndarray, weight: float, size: int = 3, radius: int = 1
) -> np.ndarray:
    """Unsharp masking using a gaussian kernel

    Args:
        img (np.ndarray): Input image.

        weight (int): Weight of the unsharp masking.

        size (int, optional): Size of the Gaussian kernel.
        Defaults to 3.

        radius (int, optional): Radius of the Gaussian kernel.
        Defaults to 1.

    Returns:
        np.ndarray: Unsharp maskied image.
    """

    return cv.addWeighted(
        img, 1, cv.GaussianBlur(img, (size, size), radius), -weight, 0
    )


def histogram_stretch_function(
    min_in: int,
    max_in: int,
    min_out: int,
    max_out: int,
    num: int = 2**8,
    range: tuple = (0, 2**8),
) -> np.ndarray:
    """Histogram stretch function.

    Args:
        a (int): Minimum input intensity.

        b (int): Maximum input intensity.

        c (int): Minimum output intensity.

        d (int): Maximum output intensity.

        num (int, optional): Number of intensity levels.
        Defaults to 2**8.

        range (tuple, optional): Range of intensity levels.
        Defaults to (0, 2**8).

    Returns:
        np.ndarray: Histogram stretch function.
    """
    input_intensity = np.linspace(range[0], range[1] - 1, num)
    output_intensity = np.concatenate(
        [
            min_in * np.ones_like(input_intensity[input_intensity < min_out]),
            np.linspace(
                min_in,
                max_in,
                np.sum((input_intensity >= min_out) & (input_intensity <= max_out)),
            ),
            max_in * np.ones_like(input_intensity[input_intensity > max_out]),
        ]
    )

    return output_intensity.astype("uint8")


def gamma_correction_function(
    gamma: float, k: int = 1, bit_depth: int = 8
) -> np.ndarray:
    """Returns the gamma correction function.

    Args:
        gamma (float): Gamma value.

        k (int, optional): Scale factor.
        Defaults to 1.

        bit (int, optional): Bit depth.
        Defaults to 8.

    Returns:
        np.ndarray: Gamma correction function.
    """
    return (
        k
        * np.power(
            np.linspace(0, 2**bit_depth - 1, 2**bit_depth) / (2**bit_depth - 1), gamma
        )
        * (2**bit_depth - 1)
    )


def log_function(i_max: int, bit: int = 8) -> np.ndarray:
    """Returns the log correction function.

    Args:
        i_max (int): Maximum image intensity.

        bit (int, optional): Bit depth.
        Defaults to 8.

    Returns:
        np.ndarray: Log correction function.
    """
    k = (2**bit - 1) / np.log(1 + i_max)
    return k * np.log(1 + np.linspace(0, 2**bit - 1, 2**bit))


def inv_log_function(i_max: int, bit: int = 8) -> np.ndarray:
    """Returns the inverse log correction funciton.

    Args:
        i_max (int): Maximum image intensity.

        bit (int, optional): Bit depth.
        Defaults to 8.

    Returns:
        np.ndarray: Inverse log correction function.
    """
    k = (2**bit - 1) / np.log(1 + i_max)
    return np.exp(np.linspace(0, 2**bit - 1, 2**bit) / k) - 1



"""
-----------------------------------------------------------------------------------------
filters module
-----------------------------------------------------------------------------------------
Provides basic spatial filters. 
"""

import numpy as np


def average_filter(k: int = 3) -> np.ndarray:
    """Returns the k x k average filter.

    Args:
        k (int): size of the filter (odd).
        Defaults to 3.

    Returns:
        np.ndarray: 'k'x'k' average filter kernel.
    """
    if k % 2 == 1:
        return np.ones((k, k)) / (k**2)
    else:
        raise (ValueError(f"k = {k} must be an odd integer."))


def gauss_filter(k: int = 3, s: float = 1) -> np.ndarray:
    """Provides the k x k Gaussian weighted average filter.
    It is calculated using the separation principle.

    Args:
        k (int, optional): Size of the kernel (odd).
        Defaults to 3.

        s (float, optional): Standard deviation.
        Defaults to 1.

    Returns:
        np.ndarray: 'k'x'k' gaussian weighted average filter kernel.
    """
    if k % 2 == 1:
        g = np.exp(-(np.arange(-(k // 2), (k // 2) + 1) ** 2) / (2 * (s**2)))[
            np.newaxis, :
        ]
        g = g / np.sum(g)

        return g.T * g
    else:
        raise (ValueError(f"k = {k} must be an odd integer."))



"""
-----------------------------------------------------------------------------------------
region module
-----------------------------------------------------------------------------------------
Provides region based segmentation algorithms. 
"""

import numpy as np
import cv2 as cv


def region_filling(
    img: np.ndarray, point: tuple, kernel: np.ndarray = None
) -> np.ndarray:
    """A simple region filling algorithm.

    Args:
        img (np.ndarray): Imput image.

        point (tuple): Starting point.

        kernel (np.ndarray, optional): Growing kernel.
        Defaults to None (Uses cross kernel).

    Returns:
        np.ndarray: Region filled image.
    """

    if kernel is None:  # Growing kernel (loading default)
        kernel = cv.getStructuringElement(cv.MORPH_CROSS, (3, 3))

    current = np.zeros_like(img)
    last = current
    last[point] = 1

    img = (img / img.max()).astype("uint8")

    noimg = np.where(img == 1, 0, 1).astype("uint8")

    current = cv.dilate(last, kernel, iterations=1) & noimg

    while (current != last).any():
        last = current
        current = cv.dilate(last, kernel, iterations=1) & noimg

    return current


def region_growing(
    img: np.ndarray,
    seed: tuple,
    percent: float,
    max_it: int = 100,
    kernel: np.ndarray = None,
) -> np.ndarray:
    """Simple region growing algorithm based on the percentage of the mean value of
    the contour pixels.

    Args:
        img (np.ndarray): Input image.

        seed (tuple): Seed.

        percent (float): Percentage threshold for the mean value.

        max_it (int): Maximum number of iteration
        Defaults to 100.

        kernel (np.ndarray, optional): Growing kernel.
        Defaults to None (Uses cross kernel).

    Returns:
        np.ndarray: Mask of the segmented image.
    """
    if kernel is not None:  # Growing kernel (loading default)
        kernel = cv.getStructuringElement(cv.MORPH_CROSS, (3, 3))

    # Initializations
    # Current Shape = seed
    current_shape = np.zeros_like(img)
    current_shape[seed] = 1

    mean_psi = img[seed]  # Value of the image at seed
    # upper and lower bound thresholds
    th = (
        np.max([0, mean_psi * (1 - percent)]).astype("uint8"),
        np.min([255, mean_psi * (1 + percent)]).astype("uint8"),
    )

    # First step
    next_shape = (
        cv.dilate(current_shape, kernel, iterations=1) - current_shape
    )  # It is just the new pixels

    selected_pxl = (img * next_shape).astype("uint8")
    selected_pxl = (selected_pxl >= th[0]) & (
        selected_pxl <= th[1]
    )  # Select those meeting criterion

    # Next steps
    i = 1
    # Stops when max iteration reached or no pixels to add
    while (selected_pxl.any()) and (i < max_it):
        # Update current shape
        current_shape = current_shape + selected_pxl.astype("uint8")

        # Determine the borders of current shape for threshold computation
        updateShape = current_shape - cv.erode(
            current_shape.astype("uint8"), kernel, iterations=1
        )

        # Update thresholds
        mean_psi = np.mean(img[updateShape == 1])
        th = (
            np.max([0, mean_psi * (1 - percent)]).astype("uint8"),
            np.min([255, mean_psi * (1 + percent)]).astype("uint8"),
        )

        # Determine next new pixels
        next_shape = (
            cv.dilate(current_shape.astype("uint8"), kernel, iterations=1)
            - current_shape
        )

        selected_pxl = (img * next_shape).astype("uint8")
        selected_pxl = (selected_pxl >= th[0]) & (
            selected_pxl <= th[1]
        )  # Select those meeting criterion

        i += 1  # counter ++

    if i == max_it:  # Warning to test more iterations
        print("Warning: Maximum iterations reached.")

    return current_shape



"""
-----------------------------------------------------------------------------------------
graph module
-----------------------------------------------------------------------------------------
Provides functionality to easily visualise images. 
"""

from collections import namedtuple
from typing import Tuple
import numpy as np
import matplotlib.pyplot as plt
import cv2 as cv


def show_image(
    img: np.ndarray,
    cmap: str = "gist_gray",
    converted: bool = True,
    normalize: bool = False,
    fig_size: tuple = (10, 20),
    fig_title: str = None,
    show_ticks: bool = False,
    x_tick: list = None,
    x_tick_labels: list = None,
    x_ticks_steps: int = 5,
    y_tick: list = None,
    y_tick_labels: list = None,
    y_ticks_steps: int = 5,
    save_figure: bool = False,
    save_name: str = "Figure.jpg",
    save_format: str = "jpg",
) -> None:
    """Show images.

    Args:
        img (np.ndarray): Image to view.

        cmap (str, optional): Colour map for the image to be displayed.
        Defaults to "gist_gray".

        converted (bool, optional): Whether an image has already been converted
        to greyscale, otherwise it will use the Open-CV API to convert it.
        Defaults to True.

        normalize (bool, optional): Whether or not to normalize an image.
        Defaults to False.

        fig_size (str, optional): Figure size in inches.
        Defaults to (10, 20).

        fig_title (str, optional): Figure title.
        Defaults to None.

        show_ticks (bool, optional): To display the x and y axis tick labels.
        Defaults to False.

        x_tick (list, optional): Values for the x-ticks when show_ticks = True.
        Defaults to None (handled by MatPlotLib).

        x_tick_labels (list, optional): Labels for the x-ticks when show_ticks = True.
        Defaults to None (handled by MatPlotLib).

        x_ticks_steps (int, optional): Steps to show x-ticks when show_ticks = True.
        Defaults to 5.

        y_tick (list, optional): Values for the y-ticks when show_ticks = True.
        Defaults to None (handled by MatPlotLib).

        y_tick_labels (list, optional): Labels for the y-ticks when show_ticks = True.
        Defaults to None (handled by MatPlotLib).

        y_ticks_steps (int, optional): Steps to show the y-ticks when show_ticks = True
        Defaults to 5.

        save_figure (bool, optional): To save the figure.
        Defaults to False.

        save_name (str, optional): Name of the output file.
        Defaults to "Figure.jpg".

        save_format (str, optional): Format of the output file.
        Defaults to "jpg".
    """

    fig = plt.figure(figsize=fig_size)
    ax = fig.gca()

    if not converted:
        img = cv.cvtColor(img, cv.COLOR_BGR2GRAY)

    if normalize:
        plt.imshow(img, cmap)
    else:
        plt.imshow(
            img, cmap, vmin=0, vmax=255
        )  # This is to avoid automatic normalization

    if show_ticks:
        x_tick = np.linspace(0, img.shape[1] - 1, x_ticks_steps, dtype=int)
        ax.set_xticks(x_tick)
        ax.set_xticklabels(
            np.linspace(0, img.shape[1] - 1, x_ticks_steps, dtype=int)
        )
        
        if x_tick is not None:  # Set the x-axis ticks
            ax.set_xticks(x_tick)
            if x_tick_labels is not None:
                ax.set_xticklabels(x_tick_labels)

        y_tick = np.linspace(0, img.shape[0] - 1, y_ticks_steps, dtype=int)
        ax.set_yticks(y_tick)
        ax.set_yticklabels(
            np.linspace(0, img.shape[0] - 1, y_ticks_steps, dtype=int)
        )
        if y_tick is not None:  # Set the y-axis ticks
            ax.set_yticks(y_tick)
            if y_tick_labels is not None:
                ax.set_yticklabels(y_tick_labels)
    else:
        ax.axis("off")

    if fig_title is not None:
        ax.set_title(fig_title)

    plt.show()

    if save_figure:
        fig.savefig(save_name, dpi=300, format=save_format, bbox_inches="tight")


def show_histogram(
    img: np.ndarray,
    total_range: tuple = (0, 2**8),
    num_bins: int = 2**8,
    ret_hist: bool = False,
    calc_cdf: bool = False,
    x_lim: tuple = None,
    y_lim: tuple = None,
    cmap: str = "gist_gray",
    save_figure: bool = False,
    save_name: str = "Figure.png",
    save_format: str = "png",
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculates and displays the histogram and CDF of an image.

    Args:
        img (np.ndarray): Input image.

        total_range (tuple, optional): Image dynamic range.
        Defaults to (0, 2**8).

        num_bins (int, optional): Number of bins for the histogram.
        Defaults to 2**8.

        ret_hist (bool, optional): Return the histogram.
        Defaults to False.

        calc_cdf (bool, optional): Calculates, visualizes and return the CDF.
        Defaults to False.

        x_lim (tuple, optional): Minimum and maximum value of the x-axis.
        Defaults to None (handled by MatPlotLib).

        y_lim (tuple, optional): Minimum and maximum value of the y-axis.
        Defaults to None (handled by MatPlotLib).

        cmap (str, optional): Colour map.
        Defaults to "gist_gray".

        save_figure (bool, optional): To save the figure.
        Defaults to False.

        save_name (str, optional): Name of the output file.
        Defaults to "Figure.jpg".

        save_format (str, optional): Format of the output file.
        Defaults to "jpg".

    Returns:
        Tuple[np.ndarray,np.ndarray]:
            histogram: Histogram of the image (only when ret_hist = True or calc_cdf = True)

            cdf: CDF of the image (only when calc_cdf = True)
    """

    histogram = cv.calcHist([img], [0], None, [num_bins], total_range)

    fig, ax = plt.subplots(2, 1, figsize=(10, 10))

    ax[0].bar(
        np.linspace(total_range[0], total_range[1] - 1, num_bins),
        histogram.flatten().astype("int"),
        width=1,
        color="k",
        label="Histogram",
    )

    if x_lim is not None:
        ax[0].set_xlim(x_lim)
        ax[0].set_xticks(
            [
                max(min(x_lim), 0),
                min((x_lim[1] - x_lim[0]) // 2, 128),
                min(max(x_lim), 255),
            ]
        )
    else:
        ax[0].set_xlim((total_range[0] - 1, total_range[1]))
        ax[0].set_xticks([0, 128, 255])

    if y_lim is not None:
        ax[0].set_ylim(y_lim)

    ax[0].set_title("Histogram", fontsize="x-large")
    ax[0].set_ylabel("Pixel Counts", fontsize="x-large")
    ax[0].tick_params(axis="both", labelsize="x-large")  # Increase size of tick labels

    if calc_cdf:
        cdf = np.cumsum(
            histogram.flatten().astype("int") / np.sum(histogram).astype("int")
        )

        ax2 = ax[0].twinx()  # clone the object

        ax2.plot(np.arange(0, 2**8), cdf, c="r", label="CDF")
        ax2.set_ylabel("Probability", fontsize="x-large")
        ax2.set_ylim((0, 1.01))

        lines, labels = ax[0].get_legend_handles_labels()
        lines2, labels2 = ax2.get_legend_handles_labels()

        ax2.legend(
            lines + lines2, labels + labels2, loc="upper left", fontsize="x-large"
        )
        ax2.tick_params(axis="y", labelsize="x-large")  # Increase size of tick labels

    # This is to show the Cmap
    ax[1].imshow(np.tile(np.arange(num_bins), (10, 1)), cmap=cmap)
    ax[1].set_xlim((-1, 256))
    box = ax[1].get_position()
    box.y0 = box.y0 + 0.19
    box.y1 = box.y1 + 0.19
    ax[1].set_position(box)
    ax[1].set_xticks([])
    ax[1].yaxis.set_visible(False)
    ax[1].set_xlabel("Intensity", fontsize="x-large")

    if save_figure:  # Saves the figure
        fig.savefig(save_name, dpi=300, format=save_format, bbox_inches="tight")

    plt.show()

    if calc_cdf and ret_hist:
        HistogramResult = namedtuple("HistogramResult", "histogram cdf")
        return HistogramResult(histogram.flatten(), cdf)
    elif ret_hist:
        HistogramResult = namedtuple("HistogramResult", "histogram")
        return HistogramResult(histogram.flatten())
